@@ -1,4 +1,4 @@
-/* $Id: s_userserv.h 22862 2006-07-08 12:04:53Z leeh $ */
+/* $Id: s_userserv.h 23875 2007-04-24 19:56:17Z leeh $ */
 #ifndef INCLUDED_s_userserv_h
 #define INCLUDED_s_userserv_h
 
@@ -8,17 +8,22 @@ struct client;
 
 struct user_reg
 {
+	unsigned int id;	/* database id */
+
 	char name[USERREGNAME_LEN];
 	char *password;
 	char *email;
 	char *suspender;
 	char *suspend_reason;
+	time_t suspend_time;
 
 	time_t reg_time;
 	time_t last_time;
 
 	int flags;
 
+	unsigned int language;
+
 	dlink_node node;
 	dlink_list channels;
 	dlink_list users;
@@ -29,10 +34,14 @@ struct user_reg
 #define US_FLAGS_SUSPENDED	0x0001
 #define US_FLAGS_PRIVATE	0x0002
 #define US_FLAGS_NEVERLOGGEDIN	0x0004
+#define US_FLAGS_NOACCESS	0x0008
+#define US_FLAGS_NOMEMOS	0x0010
 
 /* Flags not stored in the DB: 0xFFFF0000 */
 #define US_FLAGS_NEEDUPDATE	0x00010000
 
+#define USER_SUSPEND_EXPIRED(x)	((x)->flags & US_FLAGS_SUSPENDED && (x)->suspend_time && (x)->suspend_time <= CURRENT_TIME)
+
 extern struct user_reg *find_user_reg(struct client *, const char *name);
 extern struct user_reg *find_user_reg_nick(struct client *, const char *name);
 
