@@ -1,8 +1,8 @@
 /* src/client.c
  *   Contains code for handling remote clients.
  *
- * Copyright (C) 2003-2006 Lee Hardy <leeh@leeh.co.uk>
- * Copyright (C) 2003-2006 ircd-ratbox development team
+ * Copyright (C) 2003-2007 Lee Hardy <leeh@leeh.co.uk>
+ * Copyright (C) 2003-2007 ircd-ratbox development team
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -28,10 +28,11 @@
  * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  * POSSIBILITY OF SUCH DAMAGE.
  *
- * $Id: client.c 23035 2006-09-08 14:53:07Z jilles $
+ * $Id: client.c 23427 2007-01-12 20:48:17Z leeh $
  */
 #include "stdinc.h"
 #include "rserv.h"
+#include "langs.h"
 #include "client.h"
 #include "channel.h"
 #include "scommand.h"
@@ -518,6 +519,7 @@ free_client(struct client *target_p)
 {
         if(target_p->user != NULL)
 	{
+		my_free(target_p->user->ip);
 		my_free(target_p->user->mask);
                 BlockHeapFree(user_heap, target_p->user);
 	}
@@ -804,7 +806,10 @@ c_uid(struct client *client_p, const char *parv[], int parc)
 		sizeof(target_p->user->username));
 	strlcpy(target_p->user->host, parv[5], 
 		sizeof(target_p->user->host));
-	/* XXX - parv[6] == ip */
+
+	if(parv[6][0] != '0' && parv[6][1] != '\0')
+		target_p->user->ip = my_strdup(parv[6]);
+
 	strlcpy(target_p->uid, parv[7], sizeof(target_p->uid));
 	strlcpy(target_p->info, parv[8], sizeof(target_p->info));
 
