@@ -1,8 +1,8 @@
 /* src/io.c
  *   Contains code for handling input and output to sockets.
  *
- * Copyright (C) 2003-2006 Lee Hardy <leeh@leeh.co.uk>
- * Copyright (C) 2003-2006 ircd-ratbox development team
+ * Copyright (C) 2003-2007 Lee Hardy <leeh@leeh.co.uk>
+ * Copyright (C) 2003-2007 ircd-ratbox development team
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -28,7 +28,7 @@
  * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  * POSSIBILITY OF SUCH DAMAGE.
  *
- * $Id: io.c 22830 2006-06-25 18:38:32Z leeh $
+ * $Id: io.c 23499 2007-01-17 19:17:35Z androsyn $
  */
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -1487,7 +1487,7 @@ static struct addrinfo *gethostinfo(char const *host, int port)
 	hints.ai_flags = 0;
 	hints.ai_family = AF_UNSPEC;
 	hints.ai_socktype = SOCK_STREAM;
-	snprintf(portbuf, 6, "%d", port);
+	snprintf(portbuf, sizeof(portbuf), "%d", port);
 	error = getaddrinfo(host, portbuf, &hints, &res);
 	if (error)
 	{
