@@ -28,7 +28,7 @@
  * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  * POSSIBILITY OF SUCH DAMAGE.
  * 
- * $Id: service.c 23618 2007-02-13 22:14:56Z leeh $
+ * $Id: service.c 23906 2007-05-02 19:13:03Z leeh $
  */
 #include "stdinc.h"
 
@@ -36,10 +36,12 @@
 #include <crypt.h>
 #endif
 
+#include "rsdb.h"
+#include "rserv.h"
+#include "langs.h"
 #include "service.h"
 #include "client.h"
 #include "scommand.h"
-#include "rserv.h"
 #include "conf.h"
 #include "io.h"
 #include "log.h"
@@ -48,16 +50,24 @@
 #include "channel.h"
 #include "s_userserv.h"
 #include "watch.h"
+#include "balloc.h"
 
 dlink_list service_list;
+dlink_list ignore_list;
+
+static int ignore_db_callback(int, const char **);
+
+static void unmerge_service(struct client *service_p);
 
 void
 init_services(void)
 {
 	struct client *service_p;
 	dlink_node *ptr;
+	dlink_node *next_ptr;
 
-	DLINK_FOREACH(ptr, service_list.head)
+	/* init functions may remove themselves from this list */
+	DLINK_FOREACH_SAFE(ptr, next_ptr, service_list.head)
 	{
 		service_p = ptr->data;
 
@@ -67,6 +77,42 @@ init_services(void)
 		if(service_p->service->init)
 			(service_p->service->init)();
 	}
+
+	rsdb_exec(ignore_db_callback, "SELECT hostname, oper, reason FROM ignore_hosts");
+}
+
+static int
+ignore_db_callback(int argc, const char **argv)
+{
+	struct service_ignore *ignore_p;
+
+	if(EmptyString(argv[0]) || EmptyString(argv[1]) || EmptyString(argv[2]))
+		return 0;
+
+	ignore_p = my_malloc(sizeof(struct service_ignore));
+	ignore_p->mask = my_strdup(argv[0]);
+	ignore_p->oper = my_strdup(argv[1]);
+	ignore_p->reason = my_strdup(argv[2]);
+
+	dlink_add(ignore_p, &ignore_p->ptr, &ignore_list);
+	return 0;
+}
+
+static int
+find_ignore(struct client *client_p)
+{
+	struct service_ignore *ignore_p;
+	dlink_node *ptr;
+
+	DLINK_FOREACH(ptr, ignore_list.head)
+	{
+		ignore_p = ptr->data;
+
+		if(match(ignore_p->mask, client_p->user->mask))
+			return 1;
+	}
+
+	return 0;
 }
 
 typedef int (*bqcmp)(const void *, const void *);
@@ -85,50 +131,151 @@ scmd_compare(const char *name, struct service_command *cmd)
 static void
 load_service_help(struct client *service_p)
 {
-	struct service_command *scommand;
-	struct ucommand_handler *ucommand;
 	char filename[PATH_MAX];
-	dlink_node *ptr;
-	int maxlen = service_p->service->command_size / sizeof(struct service_command);
-	unsigned long i;
+	unsigned int i;
 
         if(service_p->service->command == NULL)
 		return;
 
-	scommand = service_p->service->command;
+	service_p->service->help = my_malloc(sizeof(struct cachefile *) * LANG_MAX);
+	service_p->service->helpadmin = my_malloc(sizeof(struct cachefile *) * LANG_MAX);
 
-	snprintf(filename, sizeof(filename), "%s/%s/index",
-		HELP_PATH, lcase(service_p->service->id));
+	for(i = 0; langs_available[i]; i++)
+	{
+		snprintf(filename, sizeof(filename), "%s/%s/%s/index",
+			HELP_PATH, langs_available[i], lcase(service_p->service->id));
+		service_p->service->help[i] = cache_file(filename, "index", 1);
 
-	service_p->service->help = cache_file(filename, "index");
+		strlcat(filename, "-admin", sizeof(filename));
+		service_p->service->helpadmin[i] = cache_file(filename, "index-admin", 1);
+	}
+}
 
-	snprintf(filename, sizeof(filename), "%s/%s/index-admin",
-		HELP_PATH, lcase(service_p->service->id));
+static void
+append_service_help(struct client *service_p, const char *service_id)
+{
+	struct cachefile *contents_fileptr;
+	struct cachefile *fileptr;
+	char filename[PATH_MAX];
+	unsigned int i;
 
-	service_p->service->helpadmin = cache_file(filename, "index-admin");
+	if(service_p->service->help == NULL)
+		service_p->service->help = my_malloc(sizeof(struct cachefile *) * LANG_MAX);
+
+	if(service_p->service->helpadmin == NULL)
+		service_p->service->helpadmin = my_malloc(sizeof(struct cachefile *) * LANG_MAX);
+
+	for(i = 0; langs_available[i]; i++)
+	{
+		snprintf(filename, sizeof(filename), "%s/%s/%s/index",
+			HELP_PATH, langs_available[i], lcase(service_id));
+		fileptr = cache_file(filename, "index", 1);
+
+		contents_fileptr = service_p->service->help[i];
+
+		if(contents_fileptr != NULL && fileptr != NULL)
+		{
+			dlink_move_list_tail(&fileptr->contents, &contents_fileptr->contents);
+		}
+		else if(fileptr != NULL)
+			service_p->service->help[i] = fileptr;
+
+		free_cachefile(fileptr);
+
+		strlcat(filename, "-admin", sizeof(filename));
+		fileptr = cache_file(filename, "index-admin", 1);
+
+		contents_fileptr = service_p->service->helpadmin[i];
+
+		if(contents_fileptr != NULL && fileptr != NULL)
+		{
+			/* add a blank line to the start of this file to
+			 * separate them
+			 */
+			dlink_add_alloc(emptyline, &fileptr->contents);
+			dlink_move_list_tail(&fileptr->contents, &contents_fileptr->contents);
+		}
+		else if(fileptr != NULL)
+			service_p->service->helpadmin[i] = fileptr;
+
+		free_cachefile(fileptr);
+	}
+
+}
+
+static void
+load_service_command_help(struct service_command *scommand, int maxlen,
+			struct ucommand_handler *ucommand, const char *service_id)
+{
+	char filename[PATH_MAX];
+	unsigned long i;
+	unsigned int j;
 
 	for(i = 0; i < maxlen; i++)
 	{
-		snprintf(filename, sizeof(filename), "%s/%s/",
-			HELP_PATH, lcase(service_p->service->id));
+		scommand[i].helpfile = my_malloc(sizeof(struct cachefile *) * LANG_MAX);
 
-		/* we cant lcase() twice in one function call */
-		strlcat(filename, lcase(scommand[i].cmd),
-			sizeof(filename));
+		for(j = 0; langs_available[j]; j++)
+		{
+			snprintf(filename, sizeof(filename), "%s/%s/%s/",
+				HELP_PATH, langs_available[j], lcase(service_id));
 
-		scommand[i].helpfile = cache_file(filename, scommand[i].cmd);
+			/* we cant lcase() twice in one function call */
+			strlcat(filename, lcase(scommand[i].cmd),
+				sizeof(filename));
+
+			scommand[i].helpfile[j] = cache_file(filename, scommand[i].cmd, 0);
+		}
+
+#ifdef ENABLE_USERSERV
+		/* unfortunately, userserv help on language requires extra
+		 * work to list the available languages.. do that here.
+		 */
+		if(!strcmp(service_id, "USERSERV") && !strcmp(scommand[i].cmd, "LANGUAGE"))
+		{
+			int k;
+
+			/* loop the list of langs available to update each
+			 * translation file within that
+			 */
+			for(j = 0; langs_available[j]; j++)
+			{
+				/* this doesn't have a language help file */
+				if(scommand[i].helpfile[j] == NULL)
+					continue;
+
+				/* find all translations */
+				for(k = 0; langs_available[k]; k++)
+				{
+					struct cacheline *lineptr;
+
+					if(EmptyString(langs_description[k]))
+						continue;
+
+					lineptr = BlockHeapAlloc(cacheline_heap);
+					snprintf(lineptr->data, sizeof(lineptr->data),
+						"     %-6s - %s",
+						langs_available[k], langs_description[k]);
+					dlink_add_tail(lineptr, &lineptr->linenode, &(scommand[i].helpfile[j]->contents));
+				}
+			}
+		}
+#endif
 	}
 
-	DLINK_FOREACH(ptr, service_p->service->ucommand_list.head)
+	for(i = 0; ucommand && ucommand[i].cmd && ucommand[i].cmd[0] != '\0'; i++)
 	{
-		ucommand = ptr->data;
+		ucommand[i].helpfile = my_malloc(sizeof(struct cachefile *) * LANG_MAX);
 
-	        /* now see if we can load a helpfile.. */
-        	snprintf(filename, sizeof(filename), "%s/%s/u-",
-                	 HELP_PATH, lcase(service_p->service->id));
-	        strlcat(filename, lcase(ucommand->cmd), sizeof(filename));
+		for(j = 0; langs_available[j]; j++)
+		{
+		        /* now see if we can load a helpfile.. */
+        		snprintf(filename, sizeof(filename), "%s/%s/%s/u-",
+                		 HELP_PATH, langs_available[j], lcase(service_id));
+		        strlcat(filename, lcase(ucommand->cmd), sizeof(filename));
 
-        	ucommand->helpfile = cache_file(filename, ucommand->cmd);
+	        	ucommand[i].helpfile[j] = cache_file(filename, ucommand->cmd, 0);
+		}
 	}
 }
 
@@ -137,31 +284,46 @@ clear_service_help(struct client *service_p)
 {
 	struct service_command *scommand;
 	struct ucommand_handler *ucommand;
-	dlink_node *ptr;
 	int maxlen = service_p->service->command_size / sizeof(struct service_command);
-	int i;
+	int i, j;
 	
 	if(service_p->service->command == NULL)
 		return;
 
 	scommand = service_p->service->command;
 
-	free_cachefile(service_p->service->help);
-	free_cachefile(service_p->service->helpadmin);
+	for(j = 0; langs_available[j]; j++)
+	{
+		free_cachefile(service_p->service->help[j]);
+		free_cachefile(service_p->service->helpadmin[j]);
+	}
+
+	my_free(service_p->service->help);
+	my_free(service_p->service->helpadmin);
 	service_p->service->help = NULL;
 	service_p->service->helpadmin = NULL;
 
 	for(i = 0; i < maxlen; i++)
 	{
-		free_cachefile(scommand[i].helpfile);
+		for(j = 0; langs_available[j]; j++)
+		{
+			free_cachefile(scommand[i].helpfile[j]);
+		}
+
+		my_free(scommand[i].helpfile);
 		scommand[i].helpfile = NULL;
 	}
 
-	DLINK_FOREACH(ptr, service_p->service->ucommand_list.head)
+	for(i = 0; service_p->service->ucommand && service_p->service->ucommand[i].cmd && service_p->service->ucommand[i].cmd[0] != '\0'; i++)
 	{
-		ucommand = ptr->data;
+		ucommand = &service_p->service->ucommand[i];
 
-		free_cachefile(ucommand->helpfile);
+		for(j = 0; langs_available[j]; j++)
+		{
+			free_cachefile(ucommand->helpfile[j]);
+		}
+
+		my_free(ucommand->helpfile);
 		ucommand->helpfile = NULL;
 	}
 }
@@ -169,12 +331,47 @@ clear_service_help(struct client *service_p)
 void
 rehash_help(void)
 {
+	struct service_handler *handler;
+	struct client *service_p;
 	dlink_node *ptr;
+	dlink_node *merge_ptr;
 
 	DLINK_FOREACH(ptr, service_list.head)
 	{
-		clear_service_help(ptr->data);
-		load_service_help(ptr->data);
+		service_p = ptr->data;
+
+		/* dealing with merged services makes this a little complicated.
+		 *
+		 * We clear the help when it is merged, because that can work on
+		 * the merged list of commands without issue.  The merged version
+		 * is an identical copy of the original, so they both point to the
+		 * same memory -- which we should only free() once.
+		 *
+		 * Loading the help however, requires we are looking for the directory
+		 * given by the services name.  On a merged list however, this will end
+		 * up hunting the parent service for all of the helpfiles.  To avoid
+		 * this, we "unmerge" the service back to its original state.
+		 *
+		 * Once that is done, we load the help as normal.  Then for each
+		 * service that was previously merged, we go and load the helpfiles
+		 * for it, then remerge it back up to the parent service.
+		 */
+		clear_service_help(service_p);
+		unmerge_service(service_p);
+		load_service_help(service_p);
+		load_service_command_help(service_p->service->command,
+					service_p->service->command_size / sizeof(struct service_command),
+					service_p->service->ucommand, service_p->service->id);
+
+		DLINK_FOREACH(merge_ptr, service_p->service->merged_handler_list.head)
+		{
+			handler = merge_ptr->data;
+
+			load_service_command_help(handler->command,
+						handler->command_size / sizeof(struct service_command),
+						handler->ucommand, handler->id);
+			merge_service(handler, service_p->service->id, 0);
+		}
 	}
 
 	clear_ucommand_help();
@@ -247,6 +444,9 @@ add_service(struct service_handler *service)
 
 	open_service_logfile(client_p);
 	load_service_help(client_p);
+	load_service_command_help(service->command,
+				service->command_size / sizeof(struct service_command),
+				service->ucommand, service->id);
 
 	return client_p;
 }
@@ -268,6 +468,126 @@ find_service_id(const char *name)
 	return NULL;
 }
 
+struct client *
+merge_service(struct service_handler *handler_p, const char *service_id, int startup)
+{
+	struct client *service_p;
+	struct service_command *svc_cmd;
+	struct ucommand_handler *svc_ucommand;
+	unsigned long merged_command_size;
+	unsigned int merged_command_length;
+	unsigned int original_command_length;
+
+	if((service_p = find_service_id(service_id)) == NULL)
+		return NULL;
+
+	if(startup)
+		dlink_add_tail_alloc(handler_p, &service_p->service->merged_handler_list);
+
+	/* play nice, and merge the index/index-admin helpfiles */
+	append_service_help(service_p, handler_p->id);
+
+	/* first, we do irc commands */
+	merged_command_size = service_p->service->command_size + handler_p->command_size;
+	original_command_length = service_p->service->command_size / sizeof(struct service_command);
+	merged_command_length = merged_command_size / sizeof(struct service_command);
+
+	svc_cmd = my_malloc(merged_command_size);
+	memcpy(svc_cmd, service_p->service->command, service_p->service->command_size);
+	memcpy(svc_cmd + original_command_length, handler_p->command, handler_p->command_size);
+
+	/* sort the now bigger array */
+	qsort(svc_cmd, merged_command_length,
+			sizeof(struct service_command), (bqcmp) scmd_sort);
+
+	/* we have already merged some commands into this, and we have built
+	 * the new structure via memcpy(), so we don't need the old copy anymore
+	 */
+	if(service_p->service->orig_command)
+		my_free(service_p->service->command);
+
+	if(service_p->service->orig_command == NULL)
+	{
+		service_p->service->orig_command = service_p->service->command;
+		service_p->service->orig_command_size = service_p->service->command_size;
+	}
+
+	service_p->service->command = svc_cmd;
+	service_p->service->command_size = merged_command_size;
+
+	/* now do dcc commands */
+	if(handler_p->ucommand)
+	{
+		unsigned long original_command_size;
+		unsigned long new_command_size;
+		unsigned int new_command_length;
+		int i;
+
+		original_command_length = 0;
+		new_command_length = 0;
+
+		/* always need space for a NULL entry at the end */
+		merged_command_length = 1;
+
+		for(i = 0; service_p->service->ucommand && service_p->service->ucommand[i].cmd && service_p->service->ucommand[i].cmd[0] != '\0'; i++)
+		{
+			original_command_length++;
+			merged_command_length++;
+		}
+
+		for(i = 0; handler_p->ucommand[i].cmd && handler_p->ucommand[i].cmd[0] != '\0'; i++)
+		{
+			new_command_length++;
+			merged_command_length++;
+		}
+
+		original_command_size = sizeof(struct ucommand_handler) * original_command_length;
+		new_command_size = sizeof(struct ucommand_handler) * new_command_length;
+		merged_command_size = sizeof(struct ucommand_handler) * merged_command_length;
+
+		svc_ucommand = my_malloc(merged_command_size);
+
+		if(original_command_size)
+			memcpy(svc_ucommand, service_p->service->ucommand, original_command_size);
+
+		memcpy(svc_ucommand + original_command_length, handler_p->ucommand, new_command_size);
+
+		if(service_p->service->orig_ucommand)
+			my_free(service_p->service->ucommand);
+
+		if(service_p->service->orig_ucommand == NULL)
+			service_p->service->orig_ucommand = service_p->service->ucommand;
+
+		service_p->service->ucommand = svc_ucommand;
+	}
+
+	return service_p;
+}
+
+static void
+unmerge_service(struct client *service_p)
+{
+	if(service_p->service->orig_command == NULL)
+		return;
+
+	/* allocated in ram */
+	my_free(service_p->service->command);
+	service_p->service->command = service_p->service->orig_command;
+	service_p->service->command_size = service_p->service->orig_command_size;
+
+	/* may not have merged any dcc commands */
+	if(service_p->service->orig_ucommand)
+	{
+		my_free(service_p->service->ucommand);
+		service_p->service->ucommand = service_p->service->orig_ucommand;
+	}
+
+	/* mark this service as unmerged */
+	service_p->service->orig_command = NULL;
+	service_p->service->orig_command_size = 0;
+	service_p->service->orig_ucommand = NULL;
+}
+
 void
 introduce_service(struct client *target_p)
 {
@@ -385,8 +705,9 @@ handle_service_help_index(struct client *service_p, struct client *client_p)
 	 * or theres no admin file.
 	 */
 	if(ServiceShortHelp(service_p) ||
-	   (!service_p->service->help &&
-	    (!client_p->user->oper || !service_p->service->helpadmin)))
+	   ((!service_p->service->help || lang_get_cachefile(service_p->service->help, client_p) == NULL) &&
+	    (!client_p->user->oper || 
+	     (!service_p->service->helpadmin || lang_get_cachefile(service_p->service->helpadmin, client_p) == NULL))))
 	{
 		char buf[BUFSIZE];
 		struct service_command *cmd_table;
@@ -409,20 +730,18 @@ handle_service_help_index(struct client *service_p, struct client *client_p)
 
 		if(buf[0] != '\0')
 		{
-			service_error(service_p, client_p,
-				"%s Help Index. Use HELP <command> for more information",
-				service_p->name);
-			service_error(service_p, client_p, "Topics: %s", buf);
+			service_err(service_p, client_p, SVC_HELP_INDEXINFO, service_p->name);
+			service_err(service_p, client_p, SVC_HELP_TOPICS, buf);
 		}
 		else
-			service_error(service_p, client_p, "No help is available for this service.");
+			service_err(service_p, client_p, SVC_HELP_UNAVAILABLE);
 
 		service_p->service->help_count++;
 		return;
 	}
 
 	service_p->service->flood++;
-	fileptr = service_p->service->help;
+	fileptr = lang_get_cachefile(service_p->service->help, client_p);
 
 	if(fileptr)
 	{
@@ -437,11 +756,11 @@ handle_service_help_index(struct client *service_p, struct client *client_p)
 		}
 	}
 
-	fileptr = service_p->service->helpadmin;
+	fileptr = lang_get_cachefile(service_p->service->helpadmin, client_p);
 
 	if(client_p->user->oper && fileptr)
 	{
-		service_error(service_p, client_p, "Administrator commands:");
+		service_err(service_p, client_p, SVC_HELP_INDEXADMIN);
 
 		DLINK_FOREACH(ptr, fileptr->contents.head)
 		{
@@ -465,15 +784,14 @@ handle_service_help(struct client *service_p, struct client *client_p, const cha
 		struct cacheline *lineptr;
 		dlink_node *ptr;
 
-		if(cmd_entry->helpfile == NULL ||
+		if(cmd_entry->helpfile == NULL || lang_get_cachefile(cmd_entry->helpfile, client_p) == NULL ||
 		   (cmd_entry->operonly && !is_oper(client_p)))
 		{
-			service_error(service_p, client_p,
-					"No help available on %s", arg);
+			service_err(service_p, client_p, SVC_HELP_UNAVAILABLETOPIC, arg);
 			return;
 		}
 
-		fileptr = cmd_entry->helpfile;
+		fileptr = lang_get_cachefile(cmd_entry->helpfile, client_p);
 
 		DLINK_FOREACH(ptr, fileptr->contents.head)
 		{
@@ -485,7 +803,7 @@ handle_service_help(struct client *service_p, struct client *client_p, const cha
 		service_p->service->ehelp_count++;
 	}
 	else
-		service_error(service_p, client_p, "Unknown topic '%s'", arg);
+		service_err(service_p, client_p, SVC_HELP_UNAVAILABLETOPIC, arg);
 }
 
 void
@@ -521,6 +839,14 @@ handle_service(struct client *service_p, struct client *client_p,
 	/* do flood limiting */
 	if(!client_p->user->oper)
 	{
+		/* we allow opers to traverse ignores (above), together with
+		 * any oper who is about to login
+		 */
+		if(find_ignore(client_p) && 
+			(strcasecmp(command, "OLOGIN") ||
+			 find_conf_oper(client_p->user->username, client_p->user->host, client_p->user->servername) == NULL))
+			return;
+
 		if((client_p->user->flood_time + config_file.client_flood_time) < CURRENT_TIME)
 		{
 			client_p->user->flood_time = CURRENT_TIME;
@@ -537,8 +863,7 @@ handle_service(struct client *service_p, struct client *client_p,
 		if((service_p->service->flood_max && service_p->service->flood > service_p->service->flood_max) ||
 		   client_p->user->flood_count > config_file.client_flood_max)
 		{
-			service_error(service_p, client_p, 
-					"Temporarily unable to answer query. Please try again shortly.");
+			service_err(service_p, client_p, SVC_RATELIMITEDGENERIC);
 			client_p->user->flood_count++;
 			service_p->service->paced_count++;
 			return;
@@ -553,9 +878,7 @@ handle_service(struct client *service_p, struct client *client_p,
 		client_p->user->flood_count += 1;
                 service_p->service->flood += 1;
 
-		service_error(service_p, client_p,
-				"Commands to this service must be issued via /%s instead of by name.",
-				service_p->name);
+		service_err(service_p, client_p, SVC_USECOMMANDSHORTCUT, service_p->name);
 		return;
 	}
 
@@ -568,9 +891,8 @@ handle_service(struct client *service_p, struct client *client_p,
 		if(ServiceLoginHelp(service_p) && !client_p->user->user_reg &&
 		   !client_p->user->oper && !is_oper(client_p))
 		{
-			service_error(service_p, client_p, 
-					"You must be logged in for %s::HELP",
-					service_p->name);
+			service_err(service_p, client_p, SVC_NOTLOGGEDIN,
+					service_p->name, "HELP");
 			return;
 		}
 #endif
@@ -670,7 +992,7 @@ handle_service(struct client *service_p, struct client *client_p,
 		    (!client_p->user->oper || 
 		     (client_p->user->oper->sflags & cmd_entry->operflags) == 0)))
 		{
-			service_error(service_p, client_p, "No access to %s::%s",
+			service_err(service_p, client_p, SVC_NOACCESS,
 					service_p->name, cmd_entry->cmd);
 			client_p->user->flood_count++;
 			service_p->service->flood++;
@@ -682,8 +1004,7 @@ handle_service(struct client *service_p, struct client *client_p,
 		{
 			if(client_p->user->user_reg == NULL)
 			{
-				service_error(service_p, client_p, 
-						"You must be logged in for %s::%s",
+				service_err(service_p, client_p, SVC_NOTLOGGEDIN,
 						service_p->name, cmd_entry->cmd);
 				client_p->user->flood_count++;
 				service_p->service->flood++;
@@ -699,22 +1020,32 @@ handle_service(struct client *service_p, struct client *client_p,
 
 		if(parc < cmd_entry->minparc)
 		{
-			service_error(service_p, client_p, "Insufficient parameters to %s::%s",
+			service_err(service_p, client_p, SVC_NEEDMOREPARAMS,
 					service_p->name, cmd_entry->cmd);
 			client_p->user->flood_count++;
 			service_p->service->flood++;
 			return;
 		}
 
-		retval = (cmd_entry->func)(client_p, NULL, (const char **) parv, parc);
+		cmd_entry->cmd_use++;
+
+		if(cmd_entry->func)
+			retval = (cmd_entry->func)(client_p, NULL, (const char **) parv, parc);
+		else
+			retval = 0;
+
+		/* NOTE, at this point cmd_entry may now be invalid.
+		 * Particularly if we have just done a rehash help
+		 */
+		cmd_entry = NULL;
 
 		client_p->user->flood_count += retval;
 		service_p->service->flood += retval;
-		cmd_entry->cmd_use++;
 		return;
         }
 
-        service_error(service_p, client_p, "Unknown command.");
+        service_err(service_p, client_p, SVC_UNKNOWNCOMMAND,
+			service_p->name, command);
         service_p->service->flood++;
 	client_p->user->flood_count++;
 }
@@ -738,6 +1069,25 @@ service_send(struct client *service_p, struct client *client_p,
 		sendto_one(conn_p, "%s", buf);
 }
 
+void
+service_snd(struct client *service_p, struct client *client_p,
+		struct lconn *conn_p, int msgid, ...)
+{
+	static char buf[BUFSIZE];
+	va_list args;
+
+	va_start(args, msgid);
+	vsnprintf(buf, sizeof(buf), lang_get_notice(msgid, client_p, conn_p), args);
+	va_end(args);
+
+	if(client_p)
+		sendto_server(":%s NOTICE %s :%s",
+				ServiceMsgSelf(service_p) ? SVC_UID(service_p) : MYUID, 
+				UID(client_p), buf);
+	else
+		sendto_one(conn_p, "%s", buf);
+}
+
 void
 service_error(struct client *service_p, struct client *client_p,
 		const char *format, ...)
@@ -754,6 +1104,21 @@ service_error(struct client *service_p, struct client *client_p,
 			UID(client_p), buf);
 }
 
+void
+service_err(struct client *service_p, struct client *client_p, int msgid, ...)
+{
+	static char buf[BUFSIZE];
+	va_list args;
+
+	va_start(args, msgid);
+	vsnprintf(buf, sizeof(buf), lang_get_notice(msgid, client_p, NULL), args);
+	va_end(args);
+
+	sendto_server(":%s NOTICE %s :%s",
+			ServiceMsgSelf(service_p) ? SVC_UID(service_p) : MYUID, 
+			UID(client_p), buf);
+}
+
 void
 service_stats(struct client *service_p, struct lconn *conn_p)
 {
